@@ -10,7 +10,10 @@ use Carp qw(croak);
 use Scalar::Util qw(isweak refaddr reftype);
 
 use KiokuDB::Entry;
+use KiokuDB::Entry::Skip;
 use KiokuDB::Reference;
+use KiokuDB::Collapser::Buffer;
+use KiokuDB::Error::UnknownObjects;
 
 use Data::Visitor 0.18;
 
@@ -20,10 +23,18 @@ use namespace::clean -except => 'meta';
 
 extends qw(Data::Visitor);
 
+with qw(KiokuDB::Role::UUIDs);
+
 has '+tied_as_objects' => ( default => 1 );
 
-has resolver => (
-    isa => "KiokuDB::Resolver",
+has live_objects => (
+    isa => "KiokuDB::LiveObjects",
+    is  => "ro",
+    required => 1,
+);
+
+has backend => (
+    does => "KiokuDB::Backend",
     is  => "ro",
     required => 1,
 );
@@ -31,7 +42,7 @@ has resolver => (
 has typemap_resolver => (
     isa => "KiokuDB::TypeMap::Resolver",
     is  => "ro",
-    handles => [qw(collapse_method)],
+    handles => [qw(collapse_method id_method)],
     required => 1,
 );
 
@@ -45,76 +56,13 @@ has '+weaken' => (
     default => 0,
 );
 
-has _entries => (
-    isa => 'HashRef',
+has '_buffer' => (
+    isa => "KiokuDB::Collapser::Buffer",
     is  => "ro",
-    init_arg => undef,
-    clearer => "_clear_entries",
-    writer  => "_set_entries",
+    clearer => "_clear_buffer",
+    writer  => "_set_buffer",
 );
 
-# a list of the IDs of all simple entries
-has _simple_entries => (
-    isa => 'ArrayRef',
-    is  => "ro",
-    init_arg => undef,
-    clearer => "_clear_simple_entries",
-    writer  => "_set_simple_entries",
-);
-
-# keeps track of the simple references which are first class (either weak or
-# shared, and must have an entry)
-has _first_class => (
-    isa => 'HashRef',
-    is  => "ro",
-    init_arg => undef,
-    clearer => "_clear_first_class",
-    writer  => "_set_first_class",
-);
-
-has _options => (
-    isa => 'HashRef',
-    is  => "ro",
-    init_arg => undef,
-    clearer => "_clear_options",
-    writer  => "_set_options",
-);
-
-sub clear_temp_structs {
-    my $self = shift;
-    $self->_clear_entries;
-    $self->_clear_simple_entries;
-    $self->_clear_first_class;
-    $self->_clear_options;
-}
-
-sub collapse_objects {
-    my ( $self, @objects ) = @_;
-
-    my ( $entries, @ids ) = $self->collapse( objects => \@objects );
-
-    # compute the root set
-    my @root_set = delete @{ $entries }{@ids};
-
-    # return the root set and all additional necessary entries
-    return ( @root_set, values %$entries );
-}
-
-sub collapse_known_objects {
-    my ( $self, @objects ) = @_;
-
-    my ( $entries, @ids ) = $self->collapse(
-        objects    => \@objects,
-        only_known => 1,
-    );
-
-    my @root_set = map { $_ and delete $entries->{$_} } @ids;
-
-    # return the root set and all additional necessary entries
-    # may contain undefs
-    return ( @root_set, values %$entries );
-}
-
 sub collapse {
     my ( $self, %args ) = @_;
 
@@ -122,127 +70,99 @@ sub collapse {
 
     my $r;
 
-    if ( $args{only_known} ) {
-        $args{live_objects}  ||= $self->resolver->live_objects;
-        $r = $args{resolver} ||= $args{live_objects};
-    } else {
-        $r = $args{resolver} ||= $self->resolver;
-        $args{live_objects}  ||= $args{resolver}->live_objects;
-    }
-
     if ( $args{shallow} ) {
         $args{only} = set(@$objects);
     }
 
-    my $g = Scope::Guard->new(sub {
-        $self->clear_temp_structs;
-    });
+    my $buf = KiokuDB::Collapser::Buffer->new(
+        live_objects => $self->live_objects,
+        options      => \%args,
+    );
 
-    my ( %entries, %fc );
-
-    $self->_set_entries(\%entries);
-    $self->_set_options(\%args);
-    $self->_set_first_class(\%fc);
-    $self->_set_simple_entries([]);
+    my $g = Scope::Guard->new(sub { $self->_clear_buffer });
+    $self->_set_buffer($buf);
 
     # recurse through the object, accumilating entries
     $self->visit(@$objects);
 
-    my @ids = $args{live_objects}->objects_to_ids(@$objects);
-    @fc{@ids} = ();
+    my @ids = $buf->merged_objects_to_ids(@$objects);
+
+    $buf->first_class->insert(@ids);
 
     # compact UUID space by merging simple non shared structures into a single
     # deep entry
-    $self->compact_entries() if $self->compact;
+    $buf->compact_entries if $self->compact;
 
-    return ( \%entries, @ids );
+    return ( $buf, @ids );
 }
 
-sub compact_entries {
-    my $self = shift;
+sub may_compact {
+    my ( $self, $ref_or_id ) = @_;
 
-    my ( $entries, $fc, $simple, $options ) = ( $self->_entries, $self->_first_class, $self->_simple_entries, $self->_options );
+    my $id = ref($ref_or_id) ? $ref_or_id->id : $ref_or_id;
 
-    # unify non shared simple references
-    if ( my @flatten = grep { not exists $fc->{$_} } @$simple ) {
-        my $flatten = {};
-        @{$flatten}{@flatten} = delete @{$entries}{@flatten};
-
-
-        $options->{resolver}->remove(@flatten);
-
-        $self->compact_entry($_, $flatten) for values %$entries;
-    }
-}
-
-sub compact_entry {
-    my ( $self, $entry, $flatten ) = @_;
-
-    my $data = $entry->data;
-
-    if ( $self->compact_data($data, $flatten) ) {
-        $entry->data($data);
-    }
-}
-
-sub compact_data {
-    my ( $self, $data, $flatten ) = @_;
-
-    if ( ref $data eq 'KiokuDB::Reference' ) {
-        my $id = $data->id;
-
-        if ( my $entry = $flatten->{$id} ) {
-            # replace reference with data from entry, so that the
-            # simple data is inlined, and mark that entry for removal
-            $self->compact_entry($entry, $flatten);
-
-            if ( $entry->tied or $entry->class ) {
-                $entry->clear_id;
-                $_[1] = $entry;
-            } else {
-                $_[1] = $entry->data;
-            }
-            return 1;
-        }
-    } elsif ( ref($data) eq 'ARRAY' ) {
-        ref && $self->compact_data($_, $flatten) for @$data;
-    } elsif ( ref($data) eq 'HASH' ) {
-        ref && $self->compact_data($_, $flatten) for values %$data;
-    } elsif ( ref($data) eq 'KiokuDB::Entry' ) {
-        $self->compact_entry($data, $flatten);
-    } else {
-        # passthrough
-    }
-
-    return;
+    not $self->_buffer->first_class->includes($id);
 }
 
 sub make_entry {
     my ( $self, %args ) = @_;
 
+    my $meta = delete $args{meta} || {};
+
     my $object = $args{object};
 
-    my $live_objects = $self->_options->{live_objects};
-
     if ( my $id = $args{id} ) {
-        my $prev = $live_objects->object_to_entry($object);
+        my $l = $self->live_objects;
 
-        return $self->_entries->{$id} = KiokuDB::Entry->new(
+        my $prev = $l->object_to_entry($object);
+
+        if ( !$prev and $l->id_in_storage($id) ) {
+            # FIXME Backend->store( insert => [ ... ], update => [ ... ] )
+            # this happens when keep_entries is false
+            $prev = KiokuDB::Entry->new( root => $l->id_in_root_set($id) ); # force the operation to be an update
+        }
+
+        my $entry = KiokuDB::Entry->new(
             ( $prev ? ( prev => $prev ) : () ),
             %args,
         );
+
+        $self->_buffer->insert_entry( $id => $entry, $object, %$meta );
+
+        return $entry;
     } else {
         # intrinsic
-        return KiokuDB::Entry->new(%args);
+        my $entry = KiokuDB::Entry->new(%args);
+
+        $self->_buffer->insert_intrinsic( $object => $entry, %$meta );
+
+        return $entry;
     }
 }
 
+sub make_skip_entry {
+    my ( $self, %args ) = @_;
+
+    my $object = $args{object};
+
+    my $prev = $args{prev} || $self->live_objects->object_to_entry($object);
+
+    my $id = $args{id};
+
+    unless ( $id ) {
+        croak "skip entries must have an ID" unless $prev;
+        $id = $prev->id;
+    }
+
+    return undef;
+}
+
 sub make_ref {
     my ( $self, $id, $value ) = @_;
 
     my $weak = isweak($_[2]);
 
-    $self->_first_class->{$id} = undef if $weak;
+    $self->_buffer->first_class->insert($id) if $weak;
 
     return KiokuDB::Reference->new(
         id => $id,
@@ -253,35 +173,31 @@ sub make_ref {
 sub visit_seen {
     my ( $self, $seen, $prev ) = @_;
 
-    my $id = $self->_seen_id($seen);
+    my $b = $self->_buffer;
 
+    if ( my $entry = $b->intrinsic_entry($seen) ) {
+        return $entry->clone;
+    } elsif ( my $id = $self->_buffer->object_to_id($seen) || $self->live_objects->object_to_id($seen) ) {
+        $self->_buffer->first_class->insert($id) unless blessed($seen);
 
-    # register ID as first class
-    $self->_first_class->{$id} = undef;
-
-    # return a uuid ref
-    return $self->make_ref( $id => $_[1] );
-}
-
-sub _seen_id {
-    my ( $self, $seen ) = @_;
-
-    if ( my $id = $self->_options->{live_objects}->object_to_id($seen) ) {
-        return $id;
+        # return a uuid ref
+        return $self->make_ref( $id => $_[1] );
+    } else {
+        KiokuDB::Error::UnknownObjects->throw( objects => [ $seen ] );
     }
-
-    die { unknown => $seen };
 }
 
-sub visit_ref {
+sub visit_ref_fallback {
     my ( $self, $ref ) = @_;
 
-    if ( my $entry = $self->_options->{only_new} && $self->_options->{live_objects}->object_to_entry($ref) ) {
+    my $o = $self->_buffer->options;
+
+    if ( my $entry = $o->{only_in_storage} && $self->live_objects->object_to_entry($ref) ) {
         return $self->make_ref( $entry->id => $_[1] );
     }
 
     if ( my $id = $self->_ref_id($ref) ) {
-        if ( !$self->compact and my $only = $self->_options->{only} ) {
+        if ( !$self->compact and my $only = $o->{only} ) {
             unless ( $only->contains($ref) ) {
                 return $self->make_ref( $id => $_[1] );
             }
@@ -292,7 +208,7 @@ sub visit_ref {
         if ( ref($collapsed) eq 'KiokuDB::Reference' and $collapsed->id eq $id ) {
             return $collapsed; # tied
         } else {
-            push @{ $self->_simple_entries }, $id;
+            push @{ $self->_buffer->simple_entries }, $id;
 
             $self->make_entry(
                 id     => $id,
@@ -307,7 +223,7 @@ sub visit_ref {
         # compacting it later.
         return $self->SUPER::visit_ref($_[1]);
     } else {
-        die { unknown => $ref };
+        KiokuDB::Error::UnknownObjects->throw( objects => [ $ref ] );
     }
 }
 
@@ -319,15 +235,27 @@ sub visit_ref_data {
 sub _ref_id {
     my ( $self, $ref ) = @_;
 
-    if ( my $id = $self->_options->{resolver}->object_to_id($ref) ) {
-        return $id;
-    } elsif ( $self->compact ) {
-        # if we're compacting this is not an error, we just compact in place
-        # and we generate an error if we encounter this data again in _seen_id
-        return;
-    }
+    my $l = $self->live_objects;
 
-    die { unknown => $ref };
+    if ( my $id = $l->object_to_id($ref) ) {
+        return $id;
+    } else {
+        my $b = $self->_buffer;
+
+        if ( $b->options->{only_known} ) {
+            if ( $self->compact ) {
+                # if we're compacting this is not an error, we just compact in place
+                # and we generate an error if we encounter this data again in visit_seen
+                return;
+            } else {
+                KiokuDB::Error::UnknownObjects->throw( objects => [ $ref ] );
+            }
+        } else {
+            my $id = $self->generate_uuid;
+            $b->insert( $id => $ref );
+            return $id;
+        }
+    }
 }
 
 # avoid retying, we want to get back Reference or Entry objects
@@ -342,19 +270,19 @@ sub visit_tied {
     my $tie = $self->visit($tied);
 
     if ( my $id = $self->_ref_id($ref) ) {
-        if ( !$self->compact and my $only = $self->_options->{only} ) {
+        if ( !$self->compact and my $only = $self->_buffer->options->{only} ) {
             unless ( $only->contains($ref) ) {
                 return $self->make_ref( $id => $_[1] );
             }
         }
 
-        push @{ $self->_simple_entries }, $id;
+        push @{ $self->_buffer->simple_entries }, $id;
 
         $self->make_entry(
             id     => $id,
             object => $ref,
             data   => $tie,
-            tied   => reftype($ref),
+            tied   => substr(reftype($ref), 0, 1),
         );
 
         return $self->make_ref( $id => $_[2] );
@@ -362,79 +290,101 @@ sub visit_tied {
         return $self->make_entry(
             object => $ref,
             data   => $tie,
-            tied   => reftype($ref),
+            tied   => substr(reftype($ref), 0, 1),
         );
     }
 }
 
-sub visit_object {
-    my ( $self, $object ) = @_;
+sub visit_object { shift->visit_with_typemap(@_) }
+sub visit_ref { shift->visit_with_typemap(@_) }
 
-    my $collapse = $self->collapse_method(ref $object);
+sub visit_with_typemap {
+    my ( $self, $ref ) = @_;
 
-    $self->$collapse($_[1]);
+    my $collapse = $self->collapse_method(ref $ref);
+
+    shift->$collapse(@_);
 }
 
 sub collapse_first_class {
     my ( $self, $collapse, $object, @entry_args ) = @_;
 
-    if ( my $entry = $self->_options->{only_new} && $self->_options->{live_objects}->object_to_entry($object) ) {
-        return $self->make_ref( $entry->id => $_[1] );
-    }
-
     # Data::Visitor stuff for circular refs
     $self->_register_mapping( $object, $object );
 
-    my $id = $self->_object_id($object) || return;
+    my ( $l, $b ) = ( $self->live_objects, $self->_buffer );
 
-    if ( my $only = $self->_options->{only} ) {
+    my $id = $l->object_to_id($object);
+    my $in_storage = $l->id_in_storage($id);
+
+    my $o = $b->options;
+
+    if ( $o->{only_in_storage} && $in_storage ) {
+        return $self->make_ref( $id => $_[2] );
+    }
+
+    if ( my $only = $o->{only} ) {
         unless ( $only->contains($object) ) {
-            return $self->make_ref( $id => $_[1] );
+            if ( $in_storage ) {
+                return $self->make_ref( $id => $_[2] );
+            } else {
+                KiokuDB::Error::UnknownObjects->throw( objects => [ $object ] );
+            }
         }
     }
 
-    my $class = ref $object;
+    unless ( $id ) {
+        if ( $o->{only_known} ) {
+            KiokuDB::Error::UnknownObjects->throw( objects => [ $object ] );
+        } else {
+            my $id_method = $self->id_method(ref $object);
+
+            $id = $self->$id_method($object);
+
+            if ( defined( my $conflict = $l->id_to_object($id) ) ) {
+                return $self->id_conflict( $id, $_[2], $conflict );
+            } else {
+                $b->insert( $id => $object );
+            }
+        }
+    }
 
     my @args = (
         object => $object,
         id     => $id,
-        class  => $class,
+        class  => ref($object),
         @entry_args,
     );
 
-    my $data = $self->$collapse(@args);
-
-    $self->make_entry(
-        @args,
-        data => $data,
-    );
+    $self->$collapse(@args);
 
     # we pass $_[1], an alias, so that isweak works
-    return $self->make_ref( $id => $_[1] );
+    return $self->make_ref( $id => $_[2] );
 }
 
+sub id_conflict {
+    my ( $self, $id, $object, $other ) = @_;
+
+    $self->make_skip_entry( id => $id, object => $object );
+
+    $self->_buffer->insert( $id => $object );
+
+    return $self->make_ref( $id => $_[2] );
+}
+
+
 sub collapse_intrinsic {
     my ( $self, $collapse, $object, @entry_args ) = @_;
 
     my $class = ref $object;
 
-    delete $self->{_seen}{ refaddr($object) }; # FIXME Data::Visitor ->_remove_mapping?
-
     my @args = (
         object => $object,
         class  => $class,
         @entry_args,
     );
 
-    return $self->make_entry(
-        data  => $self->$collapse(@args),
-        class => $class,
-    );
-}
-
-sub _object_id {
-    my ( $self, $object ) = @_;
-    $self->_options->{resolver}->object_to_id($object) or die { unknown => $object };
+    return $self->$collapse(@args);
 }
 
 # we don't reblass in collapse_naive
@@ -458,6 +408,8 @@ data
 
 =head1 SYNOPSIS
 
+    # mostly internal
+
 =head1 DESCRIPTION
 
 The collapser simplifies real objects into L<KiokuDB::Entry> objects to pass to
@@ -471,7 +423,8 @@ Object collapsing is detailed in L</"COLLAPSING STRATEGIES">.
 The object's data will be copied into the L<KiokuDB::Entry> with references to
 other data structures translated into L<KiokuDB::Reference> objects.
 
-Reference addresses are resolved to UIDs by L<KiokuDB::Resolver>.
+Reference addresses are mapped to unique identifiers, which are generated as
+necessary.
 
 =head2 Compacting
 
@@ -497,7 +450,7 @@ The resolver consults the typemap (L<KiokuDB::TypeMap>), and caches the results
 as keyed by C<ref $object>.
 
 The typemap contains normal entries (keyed by C<ref $object eq $class>) or isa
-entries (filtered by C<$object->isa($class)>). The rationale is that a typemap
+entries (filtered by C<< $object->isa($class) >>). The rationale is that a typemap
 entry for a superclass might not support all subclasses as well.
 
 Any strategy may be collapsed as a first class object, or intrinsicly, inside
