@@ -15,12 +15,15 @@ use Scalar::Util qw(reftype weaken);
 use Symbol qw(gensym);
 use Tie::ToObject;
 
+use KiokuDB::Error::MissingObjects;
+
 use namespace::clean -except => 'meta';
 
 has live_objects => (
     isa => "KiokuDB::LiveObjects",
     is  => "ro",
     required => 1,
+    handles => [qw(id_to_object ids_to_objects object_to_id objects_to_ids id_to_entry ids_to_entries)],
 );
 
 has backend => (
@@ -32,14 +35,43 @@ has backend => (
 has typemap_resolver => (
     isa => "KiokuDB::TypeMap::Resolver",
     is  => "ro",
-    handles => [qw(expand_method)],
+    handles => [qw(expand_method refresh_method)],
     required => 1,
 );
 
-sub register_object {
-    my ( $self, $entry, $object ) = @_;
+has queue => (
+    isa => "Bool",
+    is  => "ro",
+    default => 1,
+);
 
-    $self->live_objects->insert( $entry => $object ) if $entry->id;
+has _queue => (
+    isa => "ArrayRef",
+    is  => "ro",
+    default => sub { [] },
+);
+
+has _deferred => (
+    isa => "ArrayRef",
+    is  => "ro",
+    default => sub { [] },
+);
+
+sub register_object {
+    my ( $self, $entry, $object, @args ) = @_;
+
+    if ( my $id = $entry->id ) {
+        my $l = $self->live_objects;
+
+        $l->register_entry( $id => $entry );
+        $l->register_object( $id => $object, @args );
+
+        use Scalar::Util qw(refaddr);
+        # break cycle for passthrough objects
+        if ( ref($entry->data) and refaddr($object) == refaddr($entry->data) ) {
+            weaken($entry->{data}); # FIXME there should be a MOP way to do this
+        }
+    }
 }
 
 sub expand_objects {
@@ -55,19 +87,98 @@ sub expand_objects {
         if ( defined ( my $obj = $l->id_to_object($entry->id) ) ) {
             push @objects, $obj;
         } else {
-            push @objects, $self->expand_object($entry);
+            $self->inflate_data( $entry, \($objects[@objects]) );
         }
     }
 
+    $self->load_queue;
+
     return @objects;
 }
 
 sub expand_object {
     my ( $self, $entry ) = @_;
 
-    $self->inflate_data( $entry, \(my $data) );
+    $self->inflate_data( $entry, \(my $obj) );
 
-    $data;
+    $self->load_queue;
+
+    return $obj;
+}
+
+sub queue_ref {
+    my ( $self, $ref, $into ) = @_;
+
+    if ( $self->queue ) {
+
+        #my $b = $self->backend;
+
+        #if ( $b->can("prefetch") ) {
+        #    $b->prefetch($ref->id);
+        #}
+
+        push @{ $self->_queue }, [ $ref, $into ];
+    } else {
+        if ( ref $ref ) {
+            $$into = $self->get_or_load_object($ref->id);
+            weaken($$into) if $ref->is_weak;
+        } else {
+            $$into = $self->get_or_load_object($ref);
+        }
+    }
+}
+
+sub queue_finalizer {
+    my ( $self, @hooks ) = @_;
+
+    if ( $self->queue ) {
+        push @{ $self->_deferred }, @hooks;
+    } else {
+        foreach my $hook ( @hooks ) {
+            $self->$hook();
+        }
+    }
+}
+
+sub load_queue {
+    my $self = shift;
+
+    return unless $self->queue;
+
+    my $queue = $self->_queue;
+    my $deferred = $self->_deferred;
+
+    my @queue = @$queue;
+    my @deferred = @$deferred;
+
+    @$queue = ();
+    @$deferred = ();
+
+    if ( @queue ) {
+        my @ids;
+
+        foreach my $entry ( @queue ) {
+            my $ref = $entry->[0];
+            push @ids, ref($ref) ? $ref->id : $ref;
+        }
+
+        my @objects = $self->get_or_load_objects(@ids);
+
+        foreach my $item ( @queue ) {
+            my ( $data, $into ) = @$item;
+            my $obj = shift @objects;
+
+            $$into = $obj;
+
+            weaken $$into if ref $data and $data->is_weak;
+        }
+    }
+
+    if ( @deferred ) {
+        foreach my $item ( @deferred ) {
+            $self->$item;
+        }
+    }
 }
 
 sub inflate_data {
@@ -80,31 +191,30 @@ sub inflate_data {
     unless ( ref $data ) {
         $$into = $data;
     } elsif ( ref $data eq 'KiokuDB::Reference' ) {
-        my $id = $data->id;
-        $$into = $self->get_or_load_object($id);
-        weaken($$into) if $data->is_weak;
+        $self->queue_ref( $data, $into );
     } elsif ( ref $data eq 'KiokuDB::Entry' ) {
-        my $obj;
-
-
         if ( my $class = $data->class ) {
             my $expand_method = $self->expand_method($class);
             $$into = $self->$expand_method($data);
         } else {
-            $self->inflate_data($data->data, \$obj, $data );
+            my $obj;
+
+            $self->inflate_data($data->data, \$obj, $data);
+
+            $self->load_queue; # force vivification of $obj
 
             if ( my $tie = $data->tied ) {
-                if ( $tie eq 'HASH' ) {
+                if ( $tie eq 'H' ) {
                     tie my %h, "Tie::ToObject" => $obj;
                     $obj = \%h;
-                } elsif ( $tie eq 'ARRAY' ) {
+                } elsif ( $tie eq 'A' ) {
                     tie my @a, "Tie::ToObject" => $obj;
                     $obj = \@a;
-                } elsif ( $tie eq 'GLOB' ) {
+                } elsif ( $tie eq 'G' ) {
                     my $glob = gensym();
                     tie *$glob, "Tie::ToObject" => $obj,
                     $obj = $glob;
-                } elsif ( $tie eq 'SCALAR' ) {
+                } elsif ( $tie eq 'S' ) {
                     my $scalar;
                     tie $scalar, "Tie::ToObject" => $obj;
                     $obj = \$scalar;
@@ -115,11 +225,13 @@ sub inflate_data {
 
             $$into = $obj;
         }
+
+        $data->object($$into);
     } elsif ( ref($data) eq 'HASH' ) {
         my %targ;
         $self->register_object( $entry => \%targ ) if $entry;
         foreach my $key ( keys %$data ) {
-            $self->inflate_data( $data->{$key}, \$targ{$key});
+            $self->inflate_data( $data->{$key}, \$targ{$key} );
         }
         $$into = \%targ;
     } elsif ( ref($data) eq 'ARRAY' ) {
@@ -127,7 +239,7 @@ sub inflate_data {
         $self->register_object( $entry => \@targ ) if $entry;
         for (@$data ) {
             push @targ, undef;
-            $self->inflate_data($_, \$targ[-1]);
+            $self->inflate_data( $_, \$targ[-1] );
         }
         $$into = \@targ;
     } elsif ( ref($data) eq 'SCALAR' ) {
@@ -158,7 +270,7 @@ sub get_or_load_objects {
     my %objects;
     @objects{@ids} = $self->live_objects->ids_to_objects(@ids);
 
-    my @missing = grep { not defined $objects{$_} } @ids;
+    my @missing = grep { not defined $objects{$_} } keys %objects; # @ids may contain duplicates
 
     @objects{@missing} = $self->load_objects(@missing);
 
@@ -168,27 +280,51 @@ sub get_or_load_objects {
 sub load_objects {
     my ( $self, @ids ) = @_;
 
+    return $self->expand_objects( $self->get_or_load_entries(@ids) );
+}
+
+sub get_or_load_entries {
+    my ( $self, @ids ) = @_;
+
     my %entries;
-    @entries{@ids} = $self->live_objects->ids_to_entries(@ids);
+    @entries{@ids} = $self->ids_to_entries(@ids);
 
     if ( my @load = grep { !$entries{$_} } @ids ) {
-        #confess if @load == 1;
-        @entries{@load} = $self->backend->get(@load);
-
-        if ( my @missing = grep { !$entries{$_} } @load ) {
-            die { missing => \@missing };
-        }
-
-        $self->live_objects->insert_entries( @entries{@load} );
+        @entries{@load} = $self->load_entries(@load);
     }
 
-    return $self->expand_objects( @entries{@ids} );
+    return @entries{@ids};
 }
 
 sub load_entries {
+    my ( $self, @ids ) = @_;
+
+    my @entries = $self->backend->get(@ids);
+
+    if ( @entries != @ids or grep { !$_ } @entries ) {
+        my %entries;
+        @entries{@ids} = @entries;
+        my @missing = grep { !$entries{$_} } @ids;
+
+        KiokuDB::Error::MissingObjects->throw( ids => \@missing );
+    }
+
+    my $l = $self->live_objects;
+    foreach my $entry ( @entries ) {
+        $l->register_entry( $entry->id, $entry, in_storage => 1 );
+    }
+
+    return @entries;
+}
+
+sub register_and_expand_entries {
     my ( $self, @entries ) = @_;
 
-    $self->live_objects->insert_entries(@entries),
+    my $l = $self->live_objects;
+    foreach my $entry ( @entries ) {
+        $l->register_entry( $entry->id, $entry, in_storage => 1 );
+    }
+
     $self->expand_objects(@entries);
 }
 
@@ -202,17 +338,50 @@ sub get_or_load_object {
     }
 }
 
+sub refresh_objects {
+    my ( $self, @objects ) = @_;
+
+    $self->refresh_object($_) for @objects;
+}
+
+sub refresh_object {
+    my ( $self, $object ) = @_;
+
+    my $id = $self->object_to_id($object);
+
+    my $entry = $self->load_entry($id);
+
+    my $refresh = $self->refresh_method( $entry->class );
+
+    $self->$refresh($object, $entry);
+    $self->load_queue;
+
+    return $object;
+}
+
+sub get_or_load_entry {
+    my ( $self, $id ) = @_;
+
+    return $self->id_to_entry($id) || $self->load_entry($id);
+}
+
+sub load_entry {
+    my ( $self, $id ) = @_;
+
+    my $entry = ( $self->backend->get($id) )[0]
+        or KiokuDB::Error::MissingObjects->throw( ids => [ $id ] );
+
+    $self->live_objects->register_entry( $id => $entry, in_storage => 1 );
+
+    return $entry;
+}
+
 sub load_object {
     my ( $self, $id ) = @_;
 
-    my ( $entry ) = $self->live_objects->ids_to_entries($id);
+    my $entry = $self->get_or_load_entry($id);
 
-    unless ( $entry ) {
-        $entry = ( $self->backend->get($id) )[0] || die { missing => [ $id ] };
-        $self->live_objects->insert_entries($entry );
-    }
-
-    $self->expand_object($entry);
+    return $self->expand_object($entry);
 }
 
 __PACKAGE__->meta->make_immutable;
@@ -229,6 +398,8 @@ KiokuDB::Linker - Relinks live objects from storage entries
 
 =head1 SYNOPSIS
 
+    # mostly internal
+
 =head1 DESCRIPTION
 
 The linker reconnects entry data, recreating the connected object graph in
@@ -261,4 +432,3 @@ destroyed prematurely, but allows their use in order to avoid memory leaks.
 
 =cut
 
-
